/* exp.c -- talk to an E2 EXP port through a local serial port.
 * Ralph Corderoy, ralph@inputplus.co.uk. */

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <termios.h>
#include <time.h>
#include <sys/time.h>

#ifndef __GNUC__
#define __attribute__(a)
#endif

#define FALSE 0
#define TRUE 1

#define DIM(a) (sizeof (a) / sizeof *(a))
#define END(a) ((a) + DIM(a))

/* ------------------------------------------------------------------ */

#define ASCII_STX 2
#define ASCII_ACK 6
#define ASCII_ESC 27

/* ------------------------------------------------------------------ */

typedef enum {
    /* Are transmitting ESCs. */
    CS_TXING_ESC,
    /* Have finished transmitting ESCs because txlen seconds have
     * passed. */
    CS_TX_REQUEST_ONESHOT,
    /* Have received ACK. */
    CS_HAVE_ACK,
} comm_state;

typedef struct {
    /* Start of buffer. */
    char *buf;
    /* One byte past the end of the buffer. */
    char *end;
    /* First free byte in buffer. */
    char *free;
    /* Next byte to transmit. */
    char *tx;
} tx_buffer;

/* ------------------------------------------------------------------ */

typedef unsigned char ui8;
typedef unsigned short ui16;
typedef unsigned int ui32;

typedef signed char si8;
typedef signed short si16;
typedef signed int si32;

#define PBL_WIRE_CHAR(p, c) \
    do { \
        (*(ui8 *)(p) = (c) & 0xff) \
    } while (0)

#define PBL_WIRE_SHORT(p, s) \
    do { \
        *((ui8 *)(p)) = (s) & 0xff; \
        *((ui8 *)(p) + 1) = (s) >> 8 & 0xff; \
    } while (0)

#define PBL_WIRE_INT(p, i) \
    do { \
        *((ui8 *)(p)) = (s) & 0xff; \
        *((ui8 *)(p) + 1) = (s) >> 8 & 0xff; \
        *((ui8 *)(p) + 2) = (s) >> 16 & 0xff; \
        *((ui8 *)(p) + 3) = (s) >> 24 & 0xff; \
    } while (0)

typedef struct {
    ui16 num;
} pbl_request_00;

typedef struct {
    ui16 num;
} pbl_request_02;

typedef union {
    ui16 num;
    pbl_request_00 n00;
    pbl_request_02 n02;
} pbl_request;

typedef struct {
    ui16 num;
    ui16 seventy_two;
    ui8 major;
    ui8 minor;
    ui16 build;
    ui16 max_len;
    ui32 flags;
} pbl_reply_82;

typedef union {
    ui16 num;
    pbl_reply_82 n82;
} pbl_reply;

typedef struct {
    ui8 stx;
    ui8 compressed;
    ui16 datalen;
    union {
        pbl_request req;
        pbl_reply reply;
    } u;
    /* ui8 cksum follows u. */
} pbl_packet;

#define PBL_PACKET_SIZE(p) \
    (sizeof ((pbl_packet *)0)->stx + \
     sizeof ((pbl_packet *)0)->compressed + \
     sizeof ((pbl_packet *)0)->datalen + (p)->datalen + sizeof(ui8))

/* ------------------------------------------------------------------ */

static void receive_signal(int sig);

static int open_tty(char *tty, speed_t baud);
static void read_and_write(int fd, double txlen, double txgap);
static void calc_pbl_checksum(pbl_packet *p);
static void init_tx_buffer(tx_buffer *txb, size_t len);
static void flush_tx_buffer(tx_buffer *txb);
static void tx(tx_buffer *txb, void *d, size_t len);
static double utime(void);

static void print_termios(struct termios *t);

static speed_t name_to_speed(char *s);
static char *speed_to_name(speed_t sp);

static void *emalloc(size_t len);

static void usage(char *s) __attribute__((noreturn));
static void error(char *fmt, ...) __attribute__((noreturn))
    __attribute__((format(printf, 1, 2)));

static void print_event(char *fmt, ...)
    __attribute__((format(printf, 1, 2)));
static void print_event_va(char *fmt, va_list args);
static void print_timestamp(void);

/* ------------------------------------------------------------------ */

static char *progname;

static int signal_received;

static char default_tty[] = "/dev/ttyS0";
static char default_baud[] = "115200";
static char default_txlen[] = "30";
static char default_txgap[] = "0.005";

/* ------------------------------------------------------------------ */

int main(int argc, char *argv[])
{
    char *tty;
    char *s;
    speed_t baud;
    double txlen;
    double txgap;
    int fd;

    (progname = strrchr(*argv, '/')) ? progname++ : (progname = *argv);

    /* Ensure line-buffered, even if it's a file, for those that may be
     * monitoring its growth. */
    setvbuf(stdout, NULL, _IOLBF, 0);

    if (argc != 1) {
        usage("wrong number of arguments.");
    }

    if ((tty = getenv("EXPTTY")) == NULL) {
        tty = default_tty;
    }

    if ((s = getenv("EXPBAUD")) == NULL) {
        s = default_baud;
    }
    if ((baud = name_to_speed(s)) == B0) {
        error("unknown baud rate: \"%s\".\n", s);
    }

    if ((s = getenv("EXPTXLEN")) == NULL) {
        s = default_txlen;
    }
    txlen = atof(s);

    if ((s = getenv("EXPTXGAP")) == NULL) {
        s = default_txgap;
    }
    txgap = atof(s);

    print_event("%s, version " VERSION ", $Revision: 1.56 $.\n",
        progname);
    print_event("using %s %s %g %g\n", tty, speed_to_name(baud), txlen,
        txgap);

    if (signal(SIGINT, receive_signal) == SIG_ERR ||
        signal(SIGPIPE, receive_signal) == SIG_ERR) {
        error("couldn't install signal handler.\n");
    }

    fd = open_tty(tty, baud);

    read_and_write(fd, txlen, txgap);

    if (close(fd) == -1) {
        error("couldn't close tty: \"%s\".\n", strerror(errno));
    }

    return 0;
}

static void receive_signal(int sig)
{
    if (sig != SIGINT && sig != SIGPIPE) {
        error("signal handler called for unexpected signal %u.\n", sig);
    }

    signal_received = sig;

    return;
}

/* ------------------------------------------------------------------ */

static int open_tty(char *tty, speed_t baud)
{
    int fd;
    struct termios attr;

    if ((fd = open(tty, O_RDWR | O_NOCTTY | O_NDELAY)) == -1) {
        error("couldn't open tty %s: %s.\n", tty, strerror(errno));
    }

    memset(&attr, 0, sizeof attr);
    if (tcgetattr(fd, &attr) == -1) {
        error("couldn't get tty attributes: %s.\n", strerror(errno));
    }
    print_termios(&attr);

    attr.c_iflag &= ~(INPCK | PARMRK | ISTRIP | IXON | IXOFF | IXANY |
        BRKINT | INLCR | IGNCR | ICRNL | IUCLC | IMAXBEL);
    attr.c_iflag |= IGNPAR | IGNBRK;

    /* With OPOST clear, all other c_oflag bits are ignored and no
     * output processing takes place. */
    attr.c_oflag &= ~OPOST;

    /* BAUD environment variable, plus 8N1. */
    if (cfsetispeed(&attr, baud) == -1 ||
        cfsetospeed(&attr, baud) == -1) {
        error("couldn't set baud rate %#o: %s.\n", baud,
            strerror(errno));
    }
    attr.c_cflag &= ~(CSIZE | CSTOPB | PARENB | PARODD | HUPCL |
        CRTSCTS);
    attr.c_cflag |= CS8 | CREAD | CLOCAL;

    attr.c_lflag &= ~(ISIG | ICANON | XCASE | ECHO | ECHOE | ECHOK |
        ECHONL | IEXTEN | ECHOCTL | ECHOPRT | ECHOKE | FLUSHO | PENDIN |
        TOSTOP);
    attr.c_lflag |= NOFLSH;

    print_termios(&attr);
    if (tcsetattr(fd, TCSANOW, &attr) == -1) {
        error("couldn't set attributes: %s.\n", strerror(errno));
    }

    return fd;
}

/* ------------------------------------------------------------------ */

#define NEW_STATE(s) \
    do { \
        print_event("transitioning to state %u, %s\n", (s), #s); \
        state = (s); \
    } while (0)

static void read_and_write(int fd, double txlen, double txgap)
{
    tx_buffer txb;
    comm_state state;
    char esc;
    double firsttx;
    double lasttx;
    int n;
    ui8 buf[16];
    ui8 *s;
    ui8 *ends;

    init_tx_buffer(&txb, 1024);
    NEW_STATE(CS_TXING_ESC);
    esc = ASCII_ESC;
    tx(&txb, &esc, sizeof esc);

    firsttx = lasttx = 0;
    while (1) {
        if (signal_received) {
            print_event("received signal %u, exiting.\n",
                signal_received);
            exit(1);
        }

        switch (n = read(fd, buf, sizeof buf)) {
        case -1:
            if (errno != EAGAIN) {
                error("read error: %s.\n", strerror(errno));
            }
            break;
        case 0:
            error("unexpected eof on read.\n");
            break;
        default:
            print_timestamp();
            fputs(" i", stdout);
            for (s = buf; n--; s++) {
                printf(" %02x", *s);
                if (*s == ASCII_ACK) {
                    NEW_STATE(CS_HAVE_ACK);
                    flush_tx_buffer(&txb);
                }
            }
            putchar('\n');
            break;
        }

        if (state == CS_TXING_ESC && firsttx &&
            utime() > firsttx + txlen) {
            pbl_packet pack;

            flush_tx_buffer(&txb);

            pack.stx = ASCII_STX;
            pack.compressed = FALSE;
            pack.datalen = sizeof pack.u.req.n00;
            pack.u.req.n00.num = 0;
            calc_pbl_checksum(&pack);

            tx(&txb, &pack, PBL_PACKET_SIZE(&pack));
            tx(&txb, &pack, PBL_PACKET_SIZE(&pack));
            tx(&txb, &pack, PBL_PACKET_SIZE(&pack));

            pack.stx = ASCII_STX;
            pack.compressed = FALSE;
            pack.datalen = sizeof pack.u.req.n02;
            pack.u.req.n02.num = 2;
            calc_pbl_checksum(&pack);

            tx(&txb, &pack, PBL_PACKET_SIZE(&pack));
            tx(&txb, &pack, PBL_PACKET_SIZE(&pack));
            tx(&txb, &pack, PBL_PACKET_SIZE(&pack));
            NEW_STATE(CS_TX_REQUEST_ONESHOT);
        }

        if (state == CS_TXING_ESC && utime() > lasttx + txgap &&
            txb.tx == txb.free) {
            tx(&txb, &esc, sizeof esc);
        }

        if (txb.tx != txb.free) {
            switch (n = write(fd, txb.tx, txb.free - txb.tx)) {
            case -1:
                if (errno != EAGAIN) {
                    error("write error: %s.\n", strerror(errno));
                }
                break;
            case 0:
                error("write unexpectedly returned 0.\n");
                break;
            default:
                lasttx = utime();
                if (!firsttx) {
                    firsttx = lasttx;
                }

                print_timestamp();
                fputs(" o", stdout);
                for (s = txb.tx, ends = s + n; s < ends; s++) {
                    printf(" %02x", *s);
                }
                putchar('\n');

                txb.tx += n;
                if (txb.tx == txb.free) {
                    txb.tx = txb.free = txb.buf;
                }
                break;
            }
        }
    }

    return;
}

#undef NEW_STATE

static void calc_pbl_checksum(pbl_packet *p)
{
    int sum;
    ui8 *b, *e;

    sum = 0;
    for (b = (ui8 *)&p->datalen, e = b + sizeof p->datalen + p->datalen;
        b < e; b++) {
        sum += *b;
    }
    *e = sum;

    return;
}

static void init_tx_buffer(tx_buffer *txb, size_t len)
{
    txb->buf = emalloc(len);
    txb->end = txb->buf + len;
    flush_tx_buffer(txb);

    return;
}

static void flush_tx_buffer(tx_buffer *txb)
{
    txb->free = txb->tx = txb->buf;

    return;
}

static void tx(tx_buffer *txb, void *d, size_t len)
{
    if (txb->free + len > txb->end) {
        error("tx: buffer too full for another %u bytes.\n", len);
    }
    memcpy(txb->free, d, len);
    txb->free += len;

    return;
}

static double utime(void)
{
    struct timeval tv;

    if (gettimeofday(&tv, NULL)) {
        error("gettimeofday failed: %s.\n", strerror(errno));
    }

    return tv.tv_sec + tv.tv_usec / 1e6;
}

/* ------------------------------------------------------------------ */

#define CB(bit) \
    if ((f & (bit)) == (bit)) { \
        strcat(line, " " #bit); \
    }

#define CV(val, mask) \
    if ((f & (mask)) == (val)) { \
        strcat(line, " " #val); \
    }

static void print_termios(struct termios *t)
{
    char line[1024];
    int i;
    tcflag_t f;

    strcpy(line, "c_iflag:");
    f = t->c_iflag;
    CB(IGNBRK);
    CB(BRKINT);
    CB(IGNPAR);
    CB(PARMRK);
    CB(INPCK);
    CB(ISTRIP);
    CB(INLCR);
    CB(IGNCR);
    CB(ICRNL);
    CB(IUCLC);
    CB(IXON);
    CB(IXANY);
    CB(IXOFF);
    CB(IMAXBEL);
    print_event("%s.\n", line);

    strcpy(line, "c_oflag:");
    f = t->c_oflag;
    CB(OPOST);
    CB(OLCUC);
    CB(ONLCR);
    CB(OCRNL);
    CB(ONOCR);
    CB(ONLRET);
    CB(OFILL);
    CB(OFDEL);
    print_event("%s.\n", line);

    strcpy(line, "c_cflag:");
    f = t->c_cflag;
    CV(B0, CBAUD);
    CV(B50, CBAUD);
    CV(B75, CBAUD);
    CV(B110, CBAUD);
    CV(B134, CBAUD);
    CV(B150, CBAUD);
    CV(B200, CBAUD);
    CV(B300, CBAUD);
    CV(B600, CBAUD);
    CV(B1200, CBAUD);
    CV(B1800, CBAUD);
    CV(B2400, CBAUD);
    CV(B4800, CBAUD);
    CV(B9600, CBAUD);
    CV(B19200, CBAUD);
    CV(B38400, CBAUD);
    CV(B57600, CBAUD);
    CV(B115200, CBAUD);
    CV(B230400, CBAUD);
    CV(B460800, CBAUD);
    CV(B500000, CBAUD);
    CV(B576000, CBAUD);
    CV(B921600, CBAUD);
    CV(B1000000, CBAUD);
    CV(B1152000, CBAUD);
    CV(B1500000, CBAUD);
    CV(B2000000, CBAUD);
    CV(B2500000, CBAUD);
    CV(B3000000, CBAUD);
    CV(B3500000, CBAUD);
    CV(B4000000, CBAUD);
    CV(CS5, CSIZE);
    CV(CS6, CSIZE);
    CV(CS7, CSIZE);
    CV(CS8, CSIZE);
    CB(CSTOPB);
    CB(CREAD);
    CB(PARENB);
    CB(PARODD);
    CB(HUPCL);
    CB(CLOCAL);
    print_event("%s.\n", line);

    strcpy(line, "c_lflag:");
    f = t->c_lflag;
    CB(ISIG);
    CB(ICANON);
    CB(XCASE);
    CB(ECHO);
    CB(ECHOE);
    CB(ECHOK);
    CB(ECHONL);
    CB(NOFLSH);
    CB(TOSTOP);
    print_event("%s.\n", line);

    print_event("line: %x is: %x os: %x.\n", t->c_line, t->c_ispeed,
        t->c_ospeed);

    strcpy(line, "cc:");
    for (i = 0; i < NCCS; i++) {
        sprintf(line + strlen(line), " %x", t->c_cc[i]);
    }
    print_event("%s.\n", line);

    return;
}

#undef CB
#undef CV

/* ------------------------------------------------------------------ */

typedef struct {
    char *name;
    speed_t spd;
} speed_name;

static speed_name speeds[] = {
#define X(b) { #b, B ## b },
    X(50)
    X(75)
    X(110)
    X(134)
    X(150)
    X(200)
    X(300)
    X(600)
    X(1200)
    X(1800)
    X(2400)
    X(4800)
    X(9600)
    X(19200)
    X(38400)
    X(57600)
    X(115200)
    X(230400)
#undef X
};

static speed_t name_to_speed(char *s)
{
    speed_name *n;

    for (n = speeds; n < END(speeds); n++) {
        if (!strcmp(s, n->name)) {
            return n->spd;
        }
    }

    return B0;
}

static char *speed_to_name(speed_t sp)
{
    speed_name *n;

    for (n = speeds; n < END(speeds); n++) {
        if (n->spd == sp) {
            return n->name;
        }
    }

    return "unknown_speed";
}

/* ------------------------------------------------------------------ */

static void *emalloc(size_t len)
{
    void *p;

    if (p = malloc(len)) {
        return p;
    }

    error("failed to malloc %u bytes.\n", len);
}

/* ------------------------------------------------------------------ */

static void usage(char *s)
{
    fprintf(stderr, "%s: %s\n", progname, s);
    fprintf(stderr,
"usage: [EXPTTY=tty] [EXPBAUD=baud] [EXPTXLEN=txlen] [EXPTXGAP=txgap] %s\n"
"    tty is the serial port to use, default %s.\n"
"    baud is the baud rate to use, default %s.\n"
"    txlen is the seconds spent sending escapes, default %s.\n"
"    txgap is the seconds between sending escapes, default %s.\n",
        progname, default_tty, default_baud, default_txlen,
        default_txgap);

    exit(1);
}

static void error(char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    print_event_va(fmt, args);
    va_end(args);

    va_start(args, fmt);
    fprintf(stderr, "%s: error: ", progname);
    vfprintf(stderr, fmt, args);
    va_end(args);

    exit(1);
}

/* ------------------------------------------------------------------ */

static void print_event(char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    print_event_va(fmt, args);
    va_end(args);

    return;
}

static void print_event_va(char *fmt, va_list args)
{
    print_timestamp();
    fputs(" e ", stdout);
    vfprintf(stdout, fmt, args);

    return;
}

static void print_timestamp(void)
{
    struct timeval tv;

    if (gettimeofday(&tv, NULL)) {
        error("gettimeofday failed: %s.\n", strerror(errno));
    }
    printf("%ld.%06ld", tv.tv_sec, tv.tv_usec);

    return;
}
